@@ -1,7 +1,5 @@
 # BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
-from __future__ import absolute_import
-
 import copy
 import itertools
 
@@ -30,6 +28,7 @@ from awkward._v2.index import (
 )
 
 np = ak.nplike.NumpyMetadata.instance()
+numpy = ak.nplike.Numpy.instance()
 
 optiontypes = (IndexedOptionArray, ByteMaskedArray, BitMaskedArray, UnmaskedArray)
 listtypes = (ListOffsetArray, ListArray, RegularArray)
@@ -39,7 +38,7 @@ def broadcast_pack(inputs, isscalar):
     maxlen = -1
     for x in inputs:
         if isinstance(x, Content):
-            maxlen = max(maxlen, len(x))
+            maxlen = max(maxlen, x.length)
     if maxlen < 0:
         maxlen = 1
 
@@ -50,7 +49,11 @@ def broadcast_pack(inputs, isscalar):
             nextinputs.append(RegularArray(x.array[index], maxlen, 1))
             isscalar.append(True)
         elif isinstance(x, Content):
-            nextinputs.append(RegularArray(x, len(x), 1))
+            nextinputs.append(
+                RegularArray(
+                    x, x.length if x.nplike.known_shape else 1, 1, None, None, x.nplike
+                )
+            )
             isscalar.append(False)
         else:
             nextinputs.append(x)
@@ -59,14 +62,14 @@ def broadcast_pack(inputs, isscalar):
     return nextinputs
 
 
-def broadcast_unpack(x, isscalar):
+def broadcast_unpack(x, isscalar, nplike):
     if all(isscalar):
-        if len(x) == 0:
+        if not nplike.known_shape or x.length == 0:
             return x._getitem_nothing()._getitem_nothing()
         else:
             return x[0][0]
     else:
-        if len(x) == 0:
+        if not nplike.known_shape or x.length == 0:
             return x._getitem_nothing()
         else:
             return x[0]
@@ -80,16 +83,18 @@ def in_function(options):
 
 
 def checklength(inputs, options):
-    length = len(inputs[0])
+    length = inputs[0].length
     for x in inputs[1:]:
-        if len(x) != length:
-            raise ValueError(
-                "cannot broadcast {0} of length {1} with {2} of length {3}{4}".format(
-                    type(inputs[0]).__name__,
-                    length,
-                    type(x).__name__,
-                    len(x),
-                    in_function(options),
+        if x.length != length:
+            raise ak._v2._util.error(
+                ValueError(
+                    "cannot broadcast {} of length {} with {} of length {}{}".format(
+                        type(inputs[0]).__name__,
+                        length,
+                        type(x).__name__,
+                        x.length,
+                        in_function(options),
+                    )
                 )
             )
 
@@ -99,39 +104,43 @@ def all_same_offsets(nplike, inputs):
     for x in inputs:
         if isinstance(x, ListOffsetArray):
             if offsets is None:
-                offsets = x.offsets.to(nplike)
-            elif not nplike.array_equal(offsets, x.offsets.to(nplike)):
+                offsets = x.offsets.raw(nplike)
+            elif not nplike.index_nplike.array_equal(offsets, x.offsets.raw(nplike)):
                 return False
 
         elif isinstance(x, ListArray):
-            starts = x.starts.to(nplike)
-            stops = x.stops.to(nplike)
+            starts = x.starts.raw(nplike)
+            stops = x.stops.raw(nplike)
 
-            if not nplike.array_equal(starts[1:], stops[:-1]):
+            if not nplike.index_nplike.array_equal(starts[1:], stops[:-1]):
                 return False
 
             elif offsets is None:
-                offsets = nplike.empty(len(starts) + 1, dtype=starts.dtype)
-                if len(offsets) == 1:
+                offsets = nplike.index_nplike.empty(
+                    starts.shape[0] + 1, dtype=starts.dtype
+                )
+                if offsets.shape[0] == 1:
                     offsets[0] = 0
                 else:
                     offsets[:-1] = starts
                     offsets[-1] = stops[-1]
 
-            elif not nplike.array_equal(offsets[:-1], starts) or (
-                len(stops) != 0 and offsets[-1] != stops[-1]
+            elif not nplike.index_nplike.array_equal(offsets[:-1], starts) or (
+                stops.shape[0] != 0 and offsets[-1] != stops[-1]
             ):
                 return False
 
         elif isinstance(x, RegularArray):
             if x.size == 0:
-                my_offsets = nplike.empty(0, dtype=np.int64)
+                my_offsets = nplike.index_nplike.empty(0, dtype=np.int64)
             else:
-                my_offsets = nplike.arange(0, len(x.content), x.size)
+                my_offsets = nplike.index_nplike.arange(
+                    0, x.content.length, x.size, dtype=np.int64
+                )
 
             if offsets is None:
                 offsets = my_offsets
-            elif not nplike.array_equal(offsets, my_offsets):
+            elif not nplike.index_nplike.array_equal(offsets, my_offsets):
                 return False
 
         elif isinstance(x, Content):
@@ -171,7 +180,7 @@ def apply_step(
             for obj in inputs:
                 if isinstance(obj, Content):
                     while obj.purelist_depth < maxdepth:
-                        obj = RegularArray(obj, 1, len(obj))
+                        obj = RegularArray(obj, 1, obj.length)
                 nextinputs.append(obj)
             if any(x is not y for x, y in zip(inputs, nextinputs)):
                 return apply_step(
@@ -186,7 +195,8 @@ def apply_step(
                 )
 
     # Now all lengths must agree.
-    checklength([x for x in inputs if isinstance(x, Content)], options)
+    if nplike.known_shape:
+        checklength([x for x in inputs if isinstance(x, Content)], options)
 
     # This whole function is one big switch statement.
     def continuation():
@@ -241,67 +251,120 @@ def apply_step(
 
         # Any UnionArrays?
         elif any(isinstance(x, UnionArray) for x in inputs):
-            tagslist, numtags, length = [], [], None
-            for x in inputs:
-                if isinstance(x, UnionArray):
-                    tagslist.append(x.tags.to(nplike))
-                    numtags.append(len(x.contents))
-                    if length is None:
-                        length = len(tagslist[-1])
-                    elif length != len(tagslist[-1]):
-                        raise ValueError(
-                            "cannot broadcast UnionArray of length {0} "
-                            "with UnionArray of length {1}{2}".format(
-                                length, len(tagslist[-1]), in_function(options)
-                            )
-                        )
-
-            combos = nplike.stack(tagslist, axis=-1)
-
-            all_combos = nplike.array(
-                list(itertools.product(*[range(x) for x in numtags])),
-                dtype=[(str(i), combos.dtype) for i in range(len(tagslist))],
-            )
-
-            combos = combos.view(
-                [(str(i), combos.dtype) for i in range(len(tagslist))]
-            ).reshape(length)
-
-            tags = nplike.empty(length, dtype=np.int8)
-            index = nplike.empty(length, dtype=np.int64)
-            numoutputs, outcontents = None, []
-            for tag, combo in enumerate(all_combos):
-                mask = combos == combo
-                tags[mask] = tag
-                index[mask] = nplike.arange(nplike.count_nonzero(mask))
-                nextinputs = []
-                i = 0
+            if not nplike.known_data:
+                numtags, length = [], None
                 for x in inputs:
                     if isinstance(x, UnionArray):
-                        nextinputs.append(x[mask].project(combo[str(i)]))
-                        i += 1
-                    elif isinstance(x, Content):
-                        nextinputs.append(x[mask])
-                    else:
-                        nextinputs.append(x)
-                outcontents.append(
-                    apply_step(
-                        nplike,
-                        nextinputs,
-                        action,
-                        depth,
-                        copy.copy(depth_context),
-                        lateral_context,
-                        behavior,
-                        options,
-                    )
-                )
-                assert isinstance(outcontents[-1], tuple)
-                if numoutputs is not None:
-                    assert numoutputs == len(outcontents[-1])
-                numoutputs = len(outcontents[-1])
+                        numtags.append(len(x.contents))
+                        if length is None:
+                            length = x.tags.data.shape[0]
+                assert length is not None
 
-            assert numoutputs is not None
+                all_combos = list(itertools.product(*[range(x) for x in numtags]))
+
+                tags = nplike.index_nplike.empty(length, dtype=np.int8)
+                index = nplike.index_nplike.empty(length, dtype=np.int64)
+                numoutputs, outcontents = None, []
+                for combo in all_combos:
+                    nextinputs = []
+                    i = 0
+                    for x in inputs:
+                        if isinstance(x, UnionArray):
+                            nextinputs.append(x._contents[combo[i]])
+                            i += 1
+                        else:
+                            nextinputs.append(x)
+
+                    outcontents.append(
+                        apply_step(
+                            nplike,
+                            nextinputs,
+                            action,
+                            depth,
+                            copy.copy(depth_context),
+                            lateral_context,
+                            behavior,
+                            options,
+                        )
+                    )
+                    assert isinstance(outcontents[-1], tuple)
+                    if numoutputs is None:
+                        numoutputs = len(outcontents[-1])
+                    else:
+                        assert numoutputs == len(outcontents[-1])
+
+                assert numoutputs is not None
+
+            else:
+                tagslist, numtags, length = [], [], None
+                for x in inputs:
+                    if isinstance(x, UnionArray):
+                        tagslist.append(x.tags.raw(nplike))
+                        numtags.append(len(x.contents))
+                        if length is None:
+                            length = tagslist[-1].shape[0]
+                        elif length != tagslist[-1].shape[0]:
+                            raise ak._v2._util.error(
+                                ValueError(
+                                    "cannot broadcast UnionArray of length {} "
+                                    "with UnionArray of length {}{}".format(
+                                        length,
+                                        tagslist[-1].shape[0],
+                                        in_function(options),
+                                    )
+                                )
+                            )
+                assert length is not None
+
+                combos = nplike.index_nplike.stack(tagslist, axis=-1)
+
+                all_combos = nplike.array(
+                    list(itertools.product(*[range(x) for x in numtags])),
+                    dtype=[(str(i), combos.dtype) for i in range(len(tagslist))],
+                )
+
+                combos = combos.view(
+                    [(str(i), combos.dtype) for i in range(len(tagslist))]
+                ).reshape(length)
+
+                tags = nplike.index_nplike.empty(length, dtype=np.int8)
+                index = nplike.index_nplike.empty(length, dtype=np.int64)
+                numoutputs, outcontents = None, []
+                for tag, combo in enumerate(all_combos):
+                    mask = combos == combo
+                    tags[mask] = tag
+                    index[mask] = nplike.index_nplike.arange(
+                        nplike.count_nonzero(mask), dtype=np.int64
+                    )
+                    nextinputs = []
+                    i = 0
+                    for x in inputs:
+                        if isinstance(x, UnionArray):
+                            nextinputs.append(x[mask].project(combo[str(i)]))
+                            i += 1
+                        elif isinstance(x, Content):
+                            nextinputs.append(x[mask])
+                        else:
+                            nextinputs.append(x)
+                    outcontents.append(
+                        apply_step(
+                            nplike,
+                            nextinputs,
+                            action,
+                            depth,
+                            copy.copy(depth_context),
+                            lateral_context,
+                            behavior,
+                            options,
+                        )
+                    )
+                    assert isinstance(outcontents[-1], tuple)
+                    if numoutputs is None:
+                        numoutputs = len(outcontents[-1])
+                    else:
+                        assert numoutputs == len(outcontents[-1])
+
+                assert numoutputs is not None
 
             return tuple(
                 UnionArray(
@@ -312,36 +375,53 @@ def apply_step(
 
         # Any option-types?
         elif any(isinstance(x, optiontypes) for x in inputs):
-            mask = None
-            for x in inputs:
-                if isinstance(x, optiontypes):
-                    m = x.mask_as_bool(valid_when=False, nplike=nplike)
-                    if mask is None:
-                        mask = m
-                    else:
-                        mask = nplike.bitwise_or(mask, m, out=mask)
+            if nplike.known_data:
+                mask = None
+                for x in inputs:
+                    if isinstance(x, optiontypes):
+                        m = x.mask_as_bool(valid_when=False, nplike=nplike)
+                        if mask is None:
+                            mask = m
+                        else:
+                            mask = nplike.index_nplike.bitwise_or(mask, m, out=mask)
 
-            nextmask = Index8(mask.view(np.int8))
-            index = nplike.full(len(mask), -1, dtype=np.int64)
-            index[~mask] = nplike.arange(
-                len(mask) - nplike.count_nonzero(mask), dtype=np.int64
-            )
-            index = Index64(index)
-            if any(not isinstance(x, optiontypes) for x in inputs):
-                nextindex = nplike.arange(len(mask), dtype=np.int64)
-                nextindex[mask] = -1
-                nextindex = Index64(nextindex)
-
-            nextinputs = []
-            for x in inputs:
-                if isinstance(x, optiontypes):
-                    nextinputs.append(x.project(nextmask))
-                elif isinstance(x, Content):
-                    nextinputs.append(
-                        IndexedOptionArray(nextindex, x).project(nextmask)
+                nextmask = Index8(mask.view(np.int8))
+                index = nplike.index_nplike.full(mask.shape[0], -1, dtype=np.int64)
+                index[~mask] = nplike.index_nplike.arange(
+                    mask.shape[0] - nplike.index_nplike.count_nonzero(mask),
+                    dtype=np.int64,
+                )
+                index = Index64(index)
+                if any(not isinstance(x, optiontypes) for x in inputs):
+                    nextindex = nplike.index_nplike.arange(
+                        mask.shape[0], dtype=np.int64
                     )
-                else:
-                    nextinputs.append(x)
+                    nextindex[mask] = -1
+                    nextindex = Index64(nextindex)
+
+                nextinputs = []
+                for x in inputs:
+                    if isinstance(x, optiontypes):
+                        nextinputs.append(x.project(nextmask))
+                    elif isinstance(x, Content):
+                        nextinputs.append(
+                            IndexedOptionArray(nextindex, x).project(nextmask)
+                        )
+                    else:
+                        nextinputs.append(x)
+
+            else:
+                index = None
+                nextinputs = []
+                for x in inputs:
+                    if isinstance(x, optiontypes):
+                        index = Index64(
+                            nplike.index_nplike.empty((x.length,), np.int64)
+                        )
+                        nextinputs.append(x.content)
+                    else:
+                        nextinputs.append(x)
+                assert index is not None
 
             outcontent = apply_step(
                 nplike,
@@ -365,39 +445,62 @@ def apply_step(
                 isinstance(x, RegularArray) or not isinstance(x, listtypes)
                 for x in inputs
             ):
-                maxsize = max([x.size for x in inputs if isinstance(x, RegularArray)])
-                for x in inputs:
-                    if isinstance(x, RegularArray):
-                        if maxsize > 1 and x.size == 1:
-                            tmpindex = Index64(
-                                nplike.repeat(
-                                    nplike.arange(len(x), dtype=np.int64), maxsize
+                maxsize = max(x.size for x in inputs if isinstance(x, RegularArray))
+
+                if nplike.known_data:
+                    for x in inputs:
+                        if isinstance(x, RegularArray):
+                            if maxsize > 1 and x.size == 1:
+                                tmpindex = Index64(
+                                    nplike.index_nplike.repeat(
+                                        nplike.index_nplike.arange(
+                                            x.length, dtype=np.int64
+                                        ),
+                                        maxsize,
+                                    ),
+                                    nplike=nplike,
                                 )
-                            )
 
-                nextinputs = []
-                for x in inputs:
-
-                    if isinstance(x, RegularArray):
-                        if maxsize > 1 and x.size == 1:
-                            nextinputs.append(
-                                IndexedArray(
-                                    tmpindex, x.content[: len(x) * x.size]
-                                ).project()
-                            )
-                        elif x.size == maxsize:
-                            nextinputs.append(x.content[: len(x) * x.size])
+                    nextinputs = []
+                    for x in inputs:
+                        if isinstance(x, RegularArray):
+                            if maxsize > 1 and x.size == 1:
+                                nextinputs.append(
+                                    IndexedArray(
+                                        tmpindex, x.content[: x.length * x.size]
+                                    ).project()
+                                )
+                            elif x.size == maxsize:
+                                nextinputs.append(x.content[: x.length * x.size])
+                            else:
+                                raise ak._v2._util.error(
+                                    ValueError(
+                                        "cannot broadcast RegularArray of size "
+                                        "{} with RegularArray of size {} {}".format(
+                                            x.size, maxsize, in_function(options)
+                                        )
+                                    )
+                                )
                         else:
-                            raise ValueError(
-                                "cannot broadcast RegularArray of size "
-                                "{0} with RegularArray of size {1}{2}".format(
-                                    x.size, maxsize, in_function(options)
-                                )
-                            )
-                    else:
-                        nextinputs.append(x)
+                            nextinputs.append(x)
+
+                else:
+                    nextinputs = []
+                    for x in inputs:
+                        if isinstance(x, RegularArray):
+                            nextinputs.append(x.content)
+                        else:
+                            nextinputs.append(x)
+
+                length = None
+                for x in inputs:
+                    if isinstance(x, Content):
+                        if length is None:
+                            length = x.length
+                        elif nplike.known_shape:
+                            assert length == x.length
+                assert length is not None
 
-                maxlen = max(len(x) for x in nextinputs if isinstance(x, Content))
                 outcontent = apply_step(
                     nplike,
                     nextinputs,
@@ -409,7 +512,47 @@ def apply_step(
                     options,
                 )
                 assert isinstance(outcontent, tuple)
-                return tuple(RegularArray(x, maxsize, maxlen) for x in outcontent)
+                return tuple(RegularArray(x, maxsize, length) for x in outcontent)
+
+            elif not nplike.known_data or not nplike.known_shape:
+                offsets = None
+                nextinputs = []
+                for x in inputs:
+                    if isinstance(x, ListOffsetArray):
+                        offsets = Index64(
+                            nplike.index_nplike.empty(
+                                (x.offsets.data.shape[0],), np.int64
+                            ),
+                            nplike=nplike,
+                        )
+                        nextinputs.append(x.content)
+                    elif isinstance(x, ListArray):
+                        offsets = Index64(
+                            nplike.index_nplike.empty(
+                                (x.starts.data.shape[0] + 1,), np.int64
+                            ),
+                            nplike=nplike,
+                        )
+                        nextinputs.append(x.content)
+                    elif isinstance(x, RegularArray):
+                        nextinputs.append(x.content)
+                    else:
+                        nextinputs.append(x)
+                assert offsets is not None
+
+                outcontent = apply_step(
+                    nplike,
+                    nextinputs,
+                    action,
+                    depth + 1,
+                    copy.copy(depth_context),
+                    lateral_context,
+                    behavior,
+                    options,
+                )
+                assert isinstance(outcontent, tuple)
+
+                return tuple(ListOffsetArray(offsets, x) for x in outcontent)
 
             # Not all regular, but all same offsets?
             # Optimization: https://github.com/scikit-hep/awkward-1.0/issues/442
@@ -425,10 +568,10 @@ def apply_step(
 
                     elif isinstance(x, ListArray):
                         starts, stops = x.starts, x.stops
-                        if len(starts) == 0 or len(stops) == 0:
+                        if starts.length == 0 or stops.length == 0:
                             nextinputs.append(x.content[:0])
                         else:
-                            lencontent = nplike.max(stops)
+                            lencontent = nplike.index_nplike.max(stops)
                             nextinputs.append(x.content[:lencontent])
 
                     else:
@@ -447,13 +590,21 @@ def apply_step(
                 assert isinstance(outcontent, tuple)
 
                 if isinstance(offsets, Index):
-                    return tuple(ListOffsetArray(offsets, x) for x in outcontent)
+                    return tuple(
+                        ListOffsetArray(offsets, x).toListOffsetArray64(False)
+                        for x in outcontent
+                    )
                 elif isinstance(starts, Index) and isinstance(stops, Index):
-                    return tuple(ListArray(starts, stops, x) for x in outcontent)
+                    return tuple(
+                        ListArray(starts, stops, x).toListOffsetArray64(False)
+                        for x in outcontent
+                    )
                 else:
-                    raise AssertionError(
-                        "unexpected offsets, starts: {0}, {1}".format(
-                            type(offsets), type(starts)
+                    raise ak._v2._util.error(
+                        AssertionError(
+                            "unexpected offsets, starts: {}, {}".format(
+                                type(offsets), type(starts)
+                            )
                         )
                     )
 
@@ -495,7 +646,7 @@ def apply_step(
                     # Handle implicit left-broadcasting (non-NumPy-like broadcasting).
                     elif options["left_broadcast"] and isinstance(x, Content):
                         nextinputs.append(
-                            RegularArray(x, 1, len(x))
+                            RegularArray(x, 1, x.length)
                             ._broadcast_tooffsets64(offsets)
                             .content
                         )
@@ -519,8 +670,8 @@ def apply_step(
         # Any RecordArrays?
         elif any(isinstance(x, RecordArray) for x in inputs):
             if not options["allow_records"]:
-                raise ValueError(
-                    "cannot broadcast records{0}".format(in_function(options))
+                raise ak._v2._util.error(
+                    ValueError(f"cannot broadcast records {in_function(options)}")
                 )
 
             fields, length, istuple = None, None, True
@@ -529,21 +680,25 @@ def apply_step(
                     if fields is None:
                         fields = x.fields
                     elif set(fields) != set(x.fields):
-                        raise ValueError(
-                            "cannot broadcast records because fields don't "
-                            "match{0}:\n    {1}\n    {2}".format(
-                                in_function(options),
-                                ", ".join(sorted(fields)),
-                                ", ".join(sorted(x.fields)),
+                        raise ak._v2._util.error(
+                            ValueError(
+                                "cannot broadcast records because fields don't "
+                                "match{}:\n    {}\n    {}".format(
+                                    in_function(options),
+                                    ", ".join(sorted(fields)),
+                                    ", ".join(sorted(x.fields)),
+                                )
                             )
                         )
                     if length is None:
-                        length = len(x)
-                    elif length != len(x):
-                        raise ValueError(
-                            "cannot broadcast RecordArray of length {0} "
-                            "with RecordArray of length {1}{2}".format(
-                                length, len(x), in_function(options)
+                        length = x.length
+                    elif length != x.length:
+                        raise ak._v2._util.error(
+                            ValueError(
+                                "cannot broadcast RecordArray of length {} "
+                                "with RecordArray of length {}{}".format(
+                                    length, x.length, in_function(options)
+                                )
                             )
                         )
                     if not x.is_tuple:
@@ -575,9 +730,11 @@ def apply_step(
             )
 
         else:
-            raise ValueError(
-                "cannot broadcast: {0}{1}".format(
-                    ", ".join(repr(type(x)) for x in inputs), in_function(options)
+            raise ak._v2._util.error(
+                ValueError(
+                    "cannot broadcast: {}{}".format(
+                        ", ".join(repr(type(x)) for x in inputs), in_function(options)
+                    )
                 )
             )
 
@@ -597,7 +754,7 @@ def apply_step(
     elif result is None:
         return continuation()
     else:
-        raise AssertionError(result)
+        raise ak._v2._util.error(AssertionError(result))
 
 
 def broadcast_and_apply(
@@ -613,9 +770,10 @@ def broadcast_and_apply(
     regular_to_jagged=False,
     function_name=None,
 ):
+    nplike = ak.nplike.of(*inputs)
     isscalar = []
     out = apply_step(
-        ak.nplike.of(*inputs),
+        nplike,
         broadcast_pack(inputs, isscalar),
         action,
         0,
@@ -632,4 +790,4 @@ def broadcast_and_apply(
         },
     )
     assert isinstance(out, tuple)
-    return tuple(broadcast_unpack(x, isscalar) for x in out)
+    return tuple(broadcast_unpack(x, isscalar, nplike) for x in out)
